/*
 * parse
 */

#include <ctype.h>
#include <string.h>
#include <math.h>

#include "esp.h"
#include "parse.h"
 
static KeyWord KeyWords[] = {
"intvar", TYPE_INT + DECL_VAR, parse_decl,
"floatvar", TYPE_FLOAT + DECL_VAR, parse_decl,
"stringvar", TYPE_CHR + DECL_VAR, parse_decl,
"intconst", TYPE_INT + DECL_CONST, parse_decl,
"floatconst", TYPE_FLOAT + DECL_CONST, parse_decl,
"stringconst", TYPE_CHR + DECL_CONST, parse_decl,
"intfunc", TYPE_INT + DECL_FCN, parse_fcn,
"floatfunc", TYPE_FLOAT + DECL_FCN, parse_fcn,
"stringfunc", TYPE_CHR + DECL_FCN, parse_fcn,
"voidfunc", TYPE_VOID + DECL_FCN, parse_fcn,
NULL, 0, NULL};

static OpEnt OpTbl[] = {
"=", OP_EQ,
"<", OP_LT,
">", OP_GT,
"!=", OP_NE,
"<=", OP_LE,
">=", OP_GE,
"<-", OP_ASG,
"&", OP_AND,
"|", OP_OR,
"^", OP_XOR,
"~", OP_NOT,
"+", OP_ADD,
"-", OP_SUB,
"*", OP_MUL,
"/", OP_DIV,
"%", OP_MOD,
NULL, OP_BAD, NULL};


static char msg[BUFSIZ];
int line_number=1;

/*
 *              **** PARSE INFILE ****
 *
 * This routine is called to parse a given input source file
 * and to store the necessary structures in the given virtual machine.
 *
 * The input arguments are the virtual machine and the full path to
 * the source file.
 *
 * This routine opens the source file, and if successful, calls
 * the parse_infile_statements routine to do the actual parsing.
 *
 */
void
parse_infile(Machine *machine, char *fname)
{
	FILE *fp;

	sprintf(msg, "Parsing input file %s ...", fname);
	display_message(msg);

	if ((fp = fopen(fname, "r")) == NULL)
	{
		display_error_message("Can't open input file!");
		return;
	}

	parse_infile_statements(fp, machine);
}

/*
 *              **** PARSE INFILE STATEMENTS ****
 *
 * This routine is called to parse the given input source file
 * and store the appropriate structures in the given virtual machine.
 *
 * The input arguments for this routine is the opened file pointer to
 * the source file, and a pointer to the virtual machine.
 *
 * This routine loops until done (usually EOF on the source file)
 * It calls get_token to get the next token from the source file,
 * calls get_keyword to get the appropriate routine to process
 * the keyword, and then calls that routine.  The return code
 * from the routine indicates whether the interpreter should
 * quit compiling statements.  Usually this is because an end of file
 * condition was encountered in the source file, but could be due
 * to a severe error condition.
 */
void
parse_infile_statements(FILE *fp, Machine *machine)
{
	char buf[BUFSIZ];
	char *token;
	struct keywords *keyword;
	int running=1;

	while(running)
	{
		if ((token = get_token(fp)) == NULL) break;
		if ((keyword = get_keyword(token)) == NULL)
		{
			sprintf(buf, "Invalid keyword (%s) found - statement ignored", token);
			display_error_message(buf);
			skip_to_end_of_statement(fp);
			continue;
		}
		running = keyword->fcn(fp, machine, keyword);		/* valid keyword - process statement */
		if (running < 0) skip_to_end_of_statement(fp);		/* skip bad statement */
	}

	fclose(fp);
}

void
parse_local_decl_stmt(FILE *fp, Context *context, StmtEnt *stbl)
{ /* parse a local declarative statement */
	struct keywords *keyword;

	if ((keyword = get_keyword(stbl->name)) != NULL)
    	parse_decl_stmt(fp, context, keyword);
}

char *
get_token(FILE *fp)
{
	static char buf[BUFSIZ], *p;
	int c, cnt;

	c = get_nonblank_character(fp);

	if (c == EOF)
	{
		line_number = -1;
		return NULL;
	}
	else if (c == '#')
	{ /* comment line ignore rest of line */
		while (c != '\n' && c != EOF)
		{
			c = fgetc(fp);
		}
		line_number++;
		if (c == EOF)
		{
			line_number = -1;
			return NULL;
		}
        /* end of line reached - return first token on next line */
		return get_token(fp);
	}
	else if (c == ';')
	{ /* end of statement */
		buf[0] = ';';
		buf[1] = '\0';
		return buf;
	}
	else if (isalpha(c))
	{ /* first character is letter - get alphanumeric string */
		buf[0] = c;				/* first character in name */
		c = get_alphanumeric_string(buf+1, fp);
		if (isspace(c)) c = get_nonblank_character(fp);
		if (c == '[')
		{ /* must be array element */
			for (p=buf; *p != '\0'; p++);	/* find end of token */
			*p++ = c;			/* start of array index */
			cnt = 1;		/* one open bracket so far */
			while ((c = fgetc(fp)) != EOF)
			{
				*p++ = c;			/* copy character */
				if (c == '[') cnt++;	/* another open bracket */
				else if (c == ']') cnt--;	/* close bracket */
				if (cnt == 0) break;		/* found end of array index */
			}
			*p = '\0';			/* terminate token string */
			/* the end of this loop assumes we have read one too many chars */
			if (c != EOF) c = fgetc(fp);	/* read past array element */
		}
	}
	else if (isdigit(c))
	{ /* first character is digit - get numeric string */
		buf[0] = c;
		c = get_numeric_string(buf+1, fp);
	}
	else if (c == '"')
	{ /* character string */
		buf[0] = c;
		get_string_string(buf+1, fp);
		return buf;
	}
	else
	{ /* something else - special treatment */
		buf[0] = c;		/* this is the first character */
		buf[1] = '\0';	/* assume single character */
        buf[2] = '\0';  /* in case it is a two character token */
        c = ' ';        /* don't ungetc this character */
		switch(buf[0])
		{ /* check for compound strings (e.g., <= ) */
			case '<':
			case '>':
				c = fgetc(fp);
				if (c == '=')
				{ /* LE, GE */
					buf[1] = c;
                    c = ' '; /* don't ungetc this character */
				}
                if (buf[0] == '<' && c == '-')
                { /* assign statement */
                    buf[1] = '-';
                    c = ' '; /* don't ungetc this character */
                }
				break;
			case '!':
				c = fgetc(fp);
				if (c == '=')
				{ /* not equal */
					buf[1] = c;
                    c = ' '; /* don't ungetc this character */
				}
				break;
			case '-':	/* could be operator or negative number */
				c = fgetc(fp);
				if (isdigit(c))
				{ /* negative number */
					buf[1] = c;
                    c = get_numeric_character(fp, buf+2);
				}
				/* else just the - operator */
				break;
			/* default - unknown - just return character */
		}
	}

    /*
     * At this point, buf contains the token to be returned, and
     * c contains the character that terminated the token string.
     *
     * If c in not a white space character, we must put it back
     * in the input stream so that we can get it as the first
     * character for the next token
     */
	if (!isspace(c)) ungetc(c, fp);
	return buf;
}

int
get_nonblank_character(FILE *fp)
{ /* find next non-blank character in file stream */
	int c;

	for (c = fgetc(fp); c != EOF && isspace(c); c = fgetc(fp))
		if (c == '\n') line_number++;
	if (c == EOF) line_number = -1;
	return c;
}

int
get_alphanumeric_string(char *buf, FILE *fp)
{
	int i, c;
		
	for (i=0; i < BUFSIZ-1; i++)
	{
		c = fgetc(fp);		/* get next character */
		if (c == EOF || isspace(c)) break;	/* end of token string */
		if (!isalnum(c)) break;	/* not alphanumeric - end of token */
		*buf++ = c;				/* add character to end of string */
	}
	*buf = '\0';		/* terminate string */
	if (c == '\n') line_number++;
	else if (c == EOF) line_number = -1;
	return c;
}

/*
 *              **** GET NUMERIC STRING ****
 *
 * This routine is called to extract a numeric string from the input stream
 *
 * A numeric string is a sequence of digits with at most one decimal point
 */
int
get_numeric_string(char *buf, FILE *fp)
{
	int i, c;
    int dp=0; /* flag indicating a decimal point was encountered */

	for (i=0; i < BUFSIZ-1; i++)
	{
		c = fgetc(fp);		/* get next character */
        /* check for end of numeric string */
		if (c == EOF || isspace(c)) break;	/* end of token string */
		if (!isdigit(c) && c != '.') break;	/* not numeric or decimal point */
        if (c == '.')
        { /* found a decimal point or end of token */
            if (dp) break;  /* already found one decimal point - end of string */
            dp++;           /* remember that we found a decimal point */
        }
		*buf++ = c;				/* add character to end of string */
	}
	*buf = '\0';		/* terminate string */
	if (c == '\n') line_number++;
	else if (c == EOF) line_number = -1;
	return c;
}

void
get_string_string(char *buf, FILE *fp)
{
	int i, c;

	for (i=0; i < BUFSIZ-1; i++)
	{
		c = fgetc(fp);		/* get next character */
		if (c == '"') break;	/* closing quote - end of token */
		if (c == '\\' && i < BUFSIZ-2)
		{ /* escape character - next character is special */
			c = fgetc(fp);
			switch(c)
			{
				case 'n': c = '\n'; break;
				case 'r': c = '\r'; break;
				case 't': c = '\t'; break;
				case '"': c = '"'; break; /* double quote */
				case 'b': c = '\b'; break;
                /* if unknown sequence, it will be copied to the string */
			}
		}
		*buf++ = c;				/* add character to end of string */
	}
	*buf = '\0';		/* terminate string */
}

KeyWord *
get_keyword(char *token)
{ /* find keyword entry for matchine token */
	int i;

	for (i=0; KeyWords[i].keyword != NULL; i++)
	{ /* loop thru table looking for keyword entry */
		if (strcmp(token, KeyWords[i].keyword) == 0) return KeyWords+i;
	}
	return NULL;	/* not found */
}

int
get_integer_value(char *buf, FILE *fp)
{
   int c;

   c = get_nonblank_character(fp); /* find start of string */

   while (isdigit(c))
   { /* keep getting characters until a non-digit */
	 *buf++ = c;      /* store character into buffer */
     c = fgetc(fp); /* get next character */
   }
   *buf = '\0';      /* terminate output string */
   if (c == '\n') line_number++;
   else if (c == EOF) line_number = -1;
   return c;
}

int
get_float_value(char *buf, FILE *fp)
{
   int c, decpt=0;

   while ((c = fgetc(fp)) != EOF)
   { /* keep getting characters until a non-digit */
	 if (!isdigit(c) && c != '.') break;
	 *buf++ = c;      /* store character into buffer */
	 if (c == '.')
	 { /* decimal point - should only be one */
	   if (decpt != 0) break;
	   decpt++;
	 }
   }
   *buf = '\0';      /* terminate output string */
	if (c == '\n') line_number++;
    else if (c == EOF) line_number = -1;
   return c;
}

int
get_string_value(char *buf, FILE *fp)
{
   int c;

   c = get_nonblank_character(fp);
   if (c != '"')
   {
      display_error_message("Invalid string value");
      *buf = '\0';
	  if (c == '\n') line_number++;
	  return c;
   }

   while ((c = fgetc(fp)) != EOF)
   { /* keep getting characters until a non-digit */
	 if (c == '"') break;
	 *buf++ = c;      /* store character into buffer */
   }

   if (c == '"') c = get_nonblank_character(fp);
   *buf = '\0';      /* terminate output string */
	if (c == '\n') line_number++;
	else if (c == EOF) line_number = -1;
   return c;
}

void
get_line(char *buf, FILE *fp)
{
	char *p;

	fgets(buf, BUFSIZ, fp);
	for (p = buf; *p != '\0' && *p != '\n' && *p != '\r'; p++);
	*p = '\0';
}

void
skip_to_end_of_statement(FILE *fp)
{
	int c;

	skip_to_character(fp, ';');
}

void
skip_to_character(FILE *fp, int ch)
{
	int c;

	while ((c = fgetc(fp)) != EOF && c != ch);
	if (c == EOF) line_number = -1;
}

int
encode_data_type(char *type)
{
	if (strcmp(type, "int") == 0) return TYPE_INT;
	if (strcmp(type, "float") == 0) return TYPE_FLOAT;
	if (strcmp(type, "char") == 0) return TYPE_CHR;
	if (strcmp(type, "ptr") == 0) return TYPE_PTR;
	if (strcmp(type, "void") == 0) return TYPE_VOID;
	return TYPE_BAD;
}

Value *
create_value_struct(void)
{
	Value *val;

	if ((val = (Value *)malloc(sizeof(Value))) == NULL)
	{
		display_error_message("Can't malloc memory for Value");
		return NULL;
	}
	val->name = NULL;
	val->utype = DECL_BAD;
	val->type = TYPE_BAD;
	val->size = 0;
	val->val.intval = 0L;

    return val;
}

Value *
create_value_for_token(char *name)
{
	Value *val;
	char buf[BUFSIZ];

	if (isdigit(*name) ||
		((*name == '-') || (*name == '+')) && isdigit(name[1]))
	{ /* token is a number */
		val = create_value_for_number(name);
	}
	else val = create_value_for_operator(name);

	return val;
}

Value *
create_value_for_number(char *name)
{
	Value *val;

	if ((val = create_value_struct()) == NULL)
	{
		display_error_message("Can't allocate memory for Value");
		return NULL;
	}
	val->name = strdup(name);	/* name of token is value */
	val->utype = DECL_LIT;
	val->size = 1;				/* numeric literals are always scalars */
	/* integer or floating point? */
	if (strchr(name, '.') == NULL)
	{
		val->type = TYPE_INT;
		val->val.intval = atoi(name);
	}
	else
	{
		val->type = TYPE_FLOAT;
		val->val.floatval = atof(name);
	}
	return val;
}

Value *
create_value_for_operator(char *name)
{
	Value *val;
	int i;

	for (i=0; OpTbl[i].name != NULL; i++)
	{
		if (strcmp(OpTbl[i].name, name) == 0) break;
	}
	if (OpTbl[i].name == NULL)
	{
		sprintf(msg, "Unknown operator (%s)", name);
		display_error_message(msg);
		return NULL;
	}

	if ((val = create_value_struct()) == NULL)
	{
		display_error_message("Can't allocate memory for Value");
		return NULL;
	}

	val->name = strdup(name);	/* name of token is value */
	val->utype = DECL_OP;
	val->type = TYPE_OP;
	val->size = 1;
	val->val.opval = OpTbl[i].code;

	return val;
}

void
discard_value(Value *val)
{
	if (val == NULL || val->utype != DECL_TMP) return;
	free_value(val);
}

void
free_value(Value *val)
{
	if (val == NULL) return;	/* nothing to do */
	free(val->name);			/* free name */
	if (val->size > 1)			/* an array - free array */
	free(val->val.intptr);  	/* can use any flavor pointer to free memory */
}
